// CalibrationGUI.java
/**
 * Part of PC307 Group M Project: Lego RCX WebCam Mount
 * Category: PC
 *
 * Provides a GUI through which the user can interactively
 * calibrate the camera mount.
 *
 * It is expected that a calibration target be affixed to the
 * hardware, with three points A, B, C marked on, for which we know
 * the actual (yaw,pitch) positions.
 *
 * The user moves the camera using the buttons on the form, so that the
 * crosshair on the image is over each point, and confirms the point.
 * Finally the Calculate button is pressed.
 *
 * By recording the sensor values at each of the three points, and
 * comparing these sensor values to the known actual positions (in angles)
 * of each point, the calibration parameters can be calculated.
 * These calibration parameters allow the concept of angles to be used
 * with the RCX (usually through the PCAngleLayer class).
 *
 * @author Tim Malloch
 * Assistance was also provided by
 * @author Philip Hagen
 * @author Nevin Doss
 *
 * @version 1.2
 * Created on 6 October 2003, 17:13
 */


public class CalibrationGUI extends javax.swing.JFrame {


    private static final int MAX_MOVEMENT = 50;
    private static final int MIN_MOVEMENT = 5;
    private int distance = MIN_MOVEMENT;
    private static CalibrationData cd = null;
    private SensorValues sv = null;
    private Calibrator cal = null;
    private static ImagePanel camImagePanel;
    private static boolean quitrequest = false;

    /**
     * Creates a new form CalibrationGUI
     */
    public CalibrationGUI() {

        // initialise the webcam panel
        initCapture();
        
        // initialise the frame components
        initComponents();

        // try to create a Calibrator
        try {
            cal = new Calibrator();
        }
        catch (Exception e) {
            System.out.println("Could not create Calibrator object.");
            System.exit(1);
        }

        // set instructiosn text
        instructionsTextArea.setText("Line up the centre of the cross-hair with \npoint A and select the button 'Set Point A'");

        // show the frame
        show();
    }


    /**
     * Creates a new form CalibrationGUI
     */
    public CalibrationGUI(PCCommLayer pccomm) {

        // initialise the webcam panel
        initCapture();

        // initialise the frame components
        initComponents();

        // try to create a Calibrator
        try {
            cal = new Calibrator(pccomm);
        }
        catch (Exception e)
        {
            System.out.println("Could not create Calibrator object.");
            System.exit(1);
        }

        // set instructiosn text
        instructionsTextArea.setText("Line up the centre of the cross-hair with \npoint A and select the button 'Set Point A'");

        // show the frame
        show();
    }


    /**
     * This will try to set up the capture device including the positioning
     * in the frame and the image from the webcam streaming in.
     */

    private void initCapture(){
        camImagePanel = new ImagePanel();
        camImagePanel.setLayout(null);
        
        camImagePanel.setBackground(new java.awt.Color(0, 0, 0));
        camImagePanel.setMinimumSize(new java.awt.Dimension(320, 240));
        camImagePanel.setPreferredSize(new java.awt.Dimension(320, 240));
        getContentPane().add(camImagePanel);
        camImagePanel.setBounds(0, 0, 320, 240);
        camImagePanel.setCrossHair(true);
    }
    
    /** 
     * This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        movementPanel = new javax.swing.JPanel();
        upButton = new javax.swing.JButton();
        leftButton = new javax.swing.JButton();
        rightButton = new javax.swing.JButton();
        downButton = new javax.swing.JButton();
        amountScrollBar = new javax.swing.JScrollBar();
        jLabel1 = new javax.swing.JLabel();
        jLabel11 = new javax.swing.JLabel();
        jLabel2 = new javax.swing.JLabel();
        exitButton = new javax.swing.JButton();
        instructionsPanel = new javax.swing.JPanel();
        instructionsTextArea = new javax.swing.JTextArea();
        calibrationPanel = new javax.swing.JPanel();
        setPointAButton = new javax.swing.JButton();
        setPointBButton = new javax.swing.JButton();
        setPointCButton = new javax.swing.JButton();
        calculateButton = new javax.swing.JButton();

        getContentPane().setLayout(null);

        setTitle("Calibration GUI");
        setName("Calibration GUI");
        
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                exitForm(evt);
            }
        });

        movementPanel.setLayout(null);

        movementPanel.setBorder(new javax.swing.border.TitledBorder("Movement"));
        upButton.setFont(new java.awt.Font("Dialog", 0, 12));
        upButton.setText("Up");
        upButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        upButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                upButtonActionPerformed(evt);
            }
        });

        movementPanel.add(upButton);
        upButton.setBounds(40, 30, 60, 30);

        leftButton.setFont(new java.awt.Font("Dialog", 0, 12));
        leftButton.setText("Left");
        leftButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        leftButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                leftButtonActionPerformed(evt);
            }
        });

        movementPanel.add(leftButton);
        leftButton.setBounds(10, 70, 50, 30);

        rightButton.setFont(new java.awt.Font("Dialog", 0, 12));
        rightButton.setText("Right");
        rightButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        rightButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                rightButtonActionPerformed(evt);
            }
        });

        movementPanel.add(rightButton);
        rightButton.setBounds(80, 70, 50, 30);

        downButton.setFont(new java.awt.Font("Dialog", 0, 12));
        downButton.setText("Down");
        downButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        downButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                downButtonActionPerformed(evt);
            }
        });

        movementPanel.add(downButton);
        downButton.setBounds(40, 110, 60, 30);

        amountScrollBar.setMinimum(MIN_MOVEMENT);
        amountScrollBar.setMaximum(MAX_MOVEMENT+10);
        amountScrollBar.setOrientation(javax.swing.JScrollBar.HORIZONTAL);
        amountScrollBar.setUnitIncrement(5);
        amountScrollBar.addAdjustmentListener(new java.awt.event.AdjustmentListener() {
            public void adjustmentValueChanged(java.awt.event.AdjustmentEvent evt) {
                amountScrollBarAdjustmentValueChanged(evt);
            }
        });

        movementPanel.add(amountScrollBar);
        amountScrollBar.setBounds(10, 190, 120, 20);

        jLabel1.setFont(new java.awt.Font("Dialog", 0, 12));
        jLabel1.setText("max");
        movementPanel.add(jLabel1);
        jLabel1.setBounds(100, 210, 30, 15);

        jLabel11.setFont(new java.awt.Font("Dialog", 0, 12));
        jLabel11.setText("min");
        movementPanel.add(jLabel11);
        jLabel11.setBounds(10, 210, 30, 15);

        jLabel2.setFont(new java.awt.Font("Dialog", 0, 12));
        jLabel2.setText("Amount");
        movementPanel.add(jLabel2);
        jLabel2.setBounds(10, 170, 60, 15);

        getContentPane().add(movementPanel);
        movementPanel.setBounds(350, 10, 140, 240);

        exitButton.setFont(new java.awt.Font("Dialog", 0, 10));
        exitButton.setText("Exit");
        exitButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        exitButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exitButtonActionPerformed(evt);
            }
        });

        getContentPane().add(exitButton);
        exitButton.setBounds(500, 10, 60, 30);

        instructionsPanel.setLayout(null);

        instructionsPanel.setBorder(new javax.swing.border.TitledBorder("Instructions"));
        instructionsTextArea.setBackground(new java.awt.Color(204, 204, 204));
        instructionsTextArea.setEditable(false);
        instructionsTextArea.setLineWrap(true);
        instructionsTextArea.setText("Welcome");
        instructionsPanel.add(instructionsTextArea);
        instructionsTextArea.setBounds(10, 20, 240, 80);

        getContentPane().add(instructionsPanel);
        instructionsPanel.setBounds(20, 260, 260, 110);

        calibrationPanel.setLayout(null);

        calibrationPanel.setBorder(new javax.swing.border.TitledBorder("Calibration"));
        setPointAButton.setFont(new java.awt.Font("Dialog", 0, 12));
        setPointAButton.setText("Set Point A");
        setPointAButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        setPointAButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setPointAButtonActionPerformed(evt);
            }
        });

        calibrationPanel.add(setPointAButton);
        setPointAButton.setBounds(10, 20, 80, 30);

        setPointBButton.setFont(new java.awt.Font("Dialog", 0, 12));
        setPointBButton.setText("Set Point B");
        setPointBButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        setPointBButton.setEnabled(false);
        setPointBButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setPointBButtonActionPerformed(evt);
            }
        });

        calibrationPanel.add(setPointBButton);
        setPointBButton.setBounds(100, 20, 70, 30);

        setPointCButton.setFont(new java.awt.Font("Dialog", 0, 12));
        setPointCButton.setText("Set Point C");
        setPointCButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        setPointCButton.setEnabled(false);
        setPointCButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setPointCButtonActionPerformed(evt);
            }
        });

        calibrationPanel.add(setPointCButton);
        setPointCButton.setBounds(180, 20, 70, 30);

        calculateButton.setFont(new java.awt.Font("Dialog", 0, 12));
        calculateButton.setText("Calculate");
        calculateButton.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        calculateButton.setEnabled(false);
        calculateButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                calculateButtonActionPerformed(evt);
            }
        });

        calibrationPanel.add(calculateButton);
        calculateButton.setBounds(70, 60, 130, 40);

        getContentPane().add(calibrationPanel);
        calibrationPanel.setBounds(290, 260, 260, 110);

        pack();

        // set the frame properties
        setSize(580, 420);
        setLocation(100,100);
        setResizable(false);

    }//GEN-END:initComponents
    
    
    /**
     * Effectively an event listener that sets and displays the distance that
     * the movement buttons will cause the mount to move by.   
	 * It executes every time the scroll bar is moved.
     *
     * @param evt the event representing the movement of the scrollbar
     */
    private void amountScrollBarAdjustmentValueChanged(java.awt.event.AdjustmentEvent evt) {//GEN-FIRST:event_amountScrollBarAdjustmentValueChanged
        instructionsTextArea.setText("");
        distance = amountScrollBar.getValue();
        //distance = ((((distance*100)/(amountScrollBar.getMaximum()-10))*MAX_MOVEMENT/100)<MIN_MOVEMENT? MIN_MOVEMENT: (((distance*100)/(amountScrollBar.getMaximum()-10))*MAX_MOVEMENT/100));
        instructionsTextArea.setText("Distance is " + distance);
    }//GEN-LAST:event_amountScrollBarAdjustmentValueChanged
    
    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * calculate the yaw and pitch ratios to convert sensor values to angles.
	 * It executes every time the Calculate button is pressed.
	 * 
	 * @param evt the event representing the Calculate button being pressed
	 */
    private void calculateButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_calculateButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("");
        try{
            cd = cal.calculateCalibration();
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
        instructionsTextArea.setText("Calibration completed successfully. The \nratios are: \n    Yaw ratio is: " + cd.getYawRatio() + "\n    Pitch ratio is: " + cd.getPitchRatio() + "\nYou may now exit.");
    }//GEN-LAST:event_calculateButtonActionPerformed
    
    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * set the yaw and pitch of point C.
	 * It executes every time the Set Point C button is pressed.
	 * 
	 * @param evt the event representing the Set Point C button being pressed
	 */
    private void setPointCButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setPointCButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("");
        try{
            sv = cal.setPointC();
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
        calculateButton.setEnabled(true);
        instructionsTextArea.setText("Point C is set at coordinates: \n\t( " + sv.getYawValue() + " , " + sv.getPitchValue() + " ) \nOnce you are sure that all the points are \ncorrectly lined up, select the button \n'Calculate'.");
    }//GEN-LAST:event_setPointCButtonActionPerformed
    
    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * set the yaw and pitch of point B.
	 * It executes every time the Set Point B button is pressed.
	 * 
	 * @param evt the event representing the Set Point B button being pressed
	 */
    private void setPointBButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setPointBButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("");
        try{
            sv = cal.setPointB();
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
        setPointCButton.setEnabled(true);
        instructionsTextArea.setText("Point B is set at coordinates: \n\t( " + sv.getYawValue() + " , " + sv.getPitchValue() + " ) \nLine up the centre of the cross-hair with \npoint C and select the button 'Set Point C'.");
    }//GEN-LAST:event_setPointBButtonActionPerformed

    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * set the yaw and pitch of point A.
	 * It executes every time the Set Point A button is pressed.
	 * 
	 * @param evt the event representing the Set Point A button being pressed
	 */
    private void setPointAButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setPointAButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("");
        try{
            sv = cal.setPointA();
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
        setPointBButton.setEnabled(true);
        instructionsTextArea.setText("Point A is set at coordinates: \n\t( " + sv.getYawValue() + " , " + sv.getPitchValue() + " ) \nLine up the centre of the cross-hair with \npoint B and select the button 'Set Point B'.");
    }//GEN-LAST:event_setPointAButtonActionPerformed
    
    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * move the mount down a pre-specified amount.
	 * It executes every time the Down button is pressed.
	 * 
	 * @param evt the event representing the Down button being pressed
	 */
    private void downButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_downButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("down");
        try{
            cal.moveDown(distance);
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
    }//GEN-LAST:event_downButtonActionPerformed
   
    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * move the mount left a pre-specified amount.
	 * It executes every time the Left button is pressed.
	 * 
	 * @param evt the event representing the Left button being pressed
	 */
    private void leftButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_leftButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("left");
        try{
            cal.moveLeft(distance);
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
    }//GEN-LAST:event_leftButtonActionPerformed
    

    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * move the mount right a pre-specified amount.
	 * It executes every time the Right button is pressed.
	 * 
	 * @param evt the event representing the Right button being pressed
	 */
    private void rightButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_rightButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("right");
        try{
            cal.moveRight(distance);
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
    }//GEN-LAST:event_rightButtonActionPerformed

    
    /**
	 * Effectively an event listener that calls a method in Calibrator to 
	 * move the mount up a pre-specified amount.
	 * It executes every time the Up button is pressed.
	 * 
	 * @param evt the event representing the Up button being pressed
	 */
    private void upButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_upButtonActionPerformed
        // Add your handling code here:
        instructionsTextArea.setText("up");
        try{
            cal.moveUp(distance);
        }catch(Exception e){
            instructionsTextArea.setText(e.getMessage());
        }
    }//GEN-LAST:event_upButtonActionPerformed
    

    /**
     * Effectively an event listener that exits the system.
     * It executes when the Exit button is pressed.
     *
     * @param evt the event representing the Exit button being pressed
     */
    private void exitButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitButtonActionPerformed
        // Add your handling code here:
        //instructionsTextArea.setText("");
        //System.exit(0);
        quit();
    }//GEN-LAST:event_exitButtonActionPerformed



    /**
     * Effectively an event listener that exits the system.
     * It executes when the X in the top corner of the frame is pressed.
     *
     * @param evt the event representing the X button being pressed
     */
    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm
        quit();
    }//GEN-LAST:event_exitForm
    
    
    
    /**
     * Causes the program to terminate.
     */
    private void quit()
    {
        instructionsTextArea.setText("");
        quitrequest = true;
        camImagePanel.quit();
        this.dispose();
    }

    /**
     *  Returns the communications layer being used to talk to the RCX by this object.
     *  @return The instance of PCCommLayer being used.
     */
    public PCCommLayer getCommLayer()
    {
        return cal.getCommLayer();
    }

    /**
     * The method through which the class is executed.
     *
     * @param args the command line arguments
     */
    public static void main(String args[]) {
        // create an instance of the class
        CalibrationGUI jf = new CalibrationGUI();
    }



    public static CalibrationData doCalibration()
    {
        // create an instance of the class
        CalibrationGUI jf = new CalibrationGUI();

        // wait for the calibration frame to terminate
        while (!quitrequest)
        {
            try
            {
                Thread.sleep(200);
            }
            catch (InterruptedException e)
            {
                // do nothing
            }
        }

        return cd;

    }


    public static CalibrationData doCalibration(PCCommLayer pccomm)
    {
        CalibrationGUI jf = new CalibrationGUI(pccomm);
        while (!quitrequest)
        {
            try
            {
                Thread.sleep(200);
            }
            catch (InterruptedException e)
            {
                // do nothing
            }
        }
        return cd;
    }


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel jLabel1;
    private javax.swing.JButton setPointBButton;
    private javax.swing.JPanel movementPanel;
    private javax.swing.JTextArea instructionsTextArea;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JButton setPointCButton;
    private javax.swing.JButton upButton;
    private javax.swing.JButton exitButton;
    private javax.swing.JLabel jLabel11;
    private javax.swing.JButton rightButton;
    private javax.swing.JButton calculateButton;
    private javax.swing.JScrollBar amountScrollBar;
    private javax.swing.JButton downButton;
    private javax.swing.JPanel instructionsPanel;
    private javax.swing.JPanel calibrationPanel;
    private javax.swing.JButton leftButton;
    private javax.swing.JButton setPointAButton;
    // End of variables declaration//GEN-END:variables

}
