//  Calibrator.java
/**
 *  Part of PC307 Group M Project: Lego RCX WebCam Mount
 *  Category: PC Control
 *
 *  This PC-side class handles calibration of the RCX.  It uses a
 *  PCCommLayer to communicate with the RCX, and has methods for
 *  setting the 3 calibration points (A, B, C), and calculating the
 *  calibration parameters from the 3 points.
 *
 *  This class is used by the CalibrationGUI class.
 *
 *  @author Philip Hagen
 *
 *  Last updated: 10 October 2003
 */

class Calibrator
{

/*
    need a calibration target that will be affixed at a known dist from the camera lens.
    has 3 marks on it, labelled A, B, C.
    A is directly in front of lens (eg. yaw=0, pitch=0)
    B is above A by a certain angle (eg. yaw=0, pitch=30)
    C is to right of A by a certain angle (eg. yaw=30, pitch=0)

    1.  tell user to affix calib. target to hardware.
    2.  tell user to move to calib point A
    3.  user moves camera around (calls moveslightly) until its over point A
    4.  user confirms camera is over point A
    5.  reset sensor values for point A to (0,0)
    6.  tell user to move to calib point B
    7.  user moves camera around (calls moveslightly) until its over point B
    8.  user confirms camera is over point B
    9.  record current sensor values for point B
    10. repeat steps 2-5 for point C
    11. work out pitch sensor difference between A and B
        divide pitch difference between A and B by known angle between A and B, and store this as pitchratio
    12. work out yaw sensor difference between A and C
        divide yaw difference between A and C by known angle between A and C, and store this as yawratio
    13. move camera back to point A  (ie. moveTo(posA - currentpos) )
    14. store/return the calibration data for use by the PC

*/

    private static final boolean verbose = false;

    private SensorValues sensorsA = null;
    private SensorValues sensorsB = null;
    private SensorValues sensorsC = null;

    private CalibrationData calibdata = null;

    //private double yawratio;
    //private double pitchratio;

    // actual known angles of the three calibration points A,B,C
    private static int POINTA_YAWANGLE = 0;
    private static int POINTA_PITCHANGLE = 0;
    private static int POINTB_YAWANGLE = 0;
    private static int POINTB_PITCHANGLE = 30;
    private static int POINTC_YAWANGLE = 30;
    private static int POINTC_PITCHANGLE = 0;

    // the communications layer used to talk to the RCX
    private PCCommLayer comm;


    /**
    * Default constructor, which creates a communications layer object.
    *
    *@exception Exception   if the underlying communications object could not be created
    */
    public Calibrator() throws Exception
    {
        // create the communications layer
        comm = new PCCommLayer();
    }

    /**
     *  Constructor to create a Calibrator that uses the given PCCommLayer.
     */
    public Calibrator(PCCommLayer pccomm)
    {
        comm = pccomm;
    }

    /**
     *  Returns the communications layer being used to talk to the RCX by this object.
     *  @return The instance of PCCommLayer being used.
     */
    public PCCommLayer getCommLayer()
    {
        return comm;
    }

    /**
     * Finalize method, called by Java garbage collector when object is about to be destroyed.
     * Close the comm layer.
     */
    protected void finalize()
    {
        comm.close();
    }


    /**
    *   Makes the camera move LEFT by a specified amount.
    *
    *   @param  dist    the desired distance to move in raw sensor values
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public void moveLeft(int dist) throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: moveLeft(" + dist + ") called.");

        try
        {
            // move left
            comm.moveBy(-dist, 0);
        }
        catch (IRException e)
        {
            throw new Exception("moveLeft() failed.");
        }
    }
 
    /**
    *   Makes the camera move RIGHT by a specified amount.
    *
    *   @param  dist    the desired distance to move in raw sensor values
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public void moveRight(int dist) throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: moveRight(" + dist + ") called.");

        try
        {
            // move right
            comm.moveBy(dist, 0);
        }
        catch (IRException e)
        {
            throw new Exception("moveRight() failed.");
        }
    }
    
    /**
    *   Makes the camera move UP by a specified amount.
    *
    *   @param  dist    the desired distance to move in raw sensor values
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public void moveUp(int dist) throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: moveUp(" + dist + ") called.");

        try
        {
            // move up
            comm.moveBy(0, dist);
        }
        catch (IRException e)
        {
            throw new Exception("moveUp() failed.");
        }
    }

    /**
    *   Makes the camera move DOWN by a specified amount.
    *
    *   @param  dist    the desired distance to move in raw sensor values
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public void moveDown(int dist) throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: moveDown(" + dist + ") called.");

        try
        {
            // move down
            comm.moveBy(0, -dist);
        }
        catch (IRException e)
        {
            throw new Exception("moveDown() failed.");
        }
    }

    /**
    *   Sets calibration point A to be the central point with respect to the vertical and horizontal planes.
    *
    *   @return     the current sensor values for the calibrated point
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public SensorValues setPointA() throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: setting calibration point A.");
        try
        {
            comm.setPosition(0, 0);
        }
        catch (IRException e)
        {
            throw new Exception("Could not reset sensor values to (0, 0).");
        }
        try
        {
            sensorsA = comm.getPosition();
        }
        catch (IRException e)
        {
            throw new Exception("Could not set calibration point A. Communication with RCX failed.");
        }

        if (verbose)
            System.out.println("Point A sensor values: (" + sensorsA.getYawValue() + ", " + sensorsA.getPitchValue() + ")");

        return sensorsA;
    }

    /**
    *   Gets the sensor value coordinates from point B on the calibration card,
    *   then sets calibration point B with these new sensor values.
    *
    *   @return     the current sensor values for the calibrated point
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public SensorValues setPointB() throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: setting calibration point B.");

        try
        {
            sensorsB = comm.getPosition();
        }
        catch (IRException e)
        {
            throw new Exception("Could not set calibration point B. Communication with RCX failed.");
        }

        if (verbose)
            System.out.println("Point B sensor values: (" + sensorsB.getYawValue() + ", " + sensorsB.getPitchValue() + ")");

        return sensorsB;
    }

    /**
    *   Gets the sensor value coordinates from point C on the calibration card,
    *   then sets calibration point C with these new sensor values.
    *
    *   @return     the current sensor values for the calibrated point
    *   @exception Exception   if the underlying communications object failed to execute the method call
    */
    public SensorValues setPointC() throws Exception
    {
        if (verbose)
            System.out.println("Calibrator: setting calibration point C.");

        try
        {
            sensorsC = comm.getPosition();
        }
        catch (IRException e)
        {
            throw new Exception("Could not set calibration point C. Communication with RCX failed.");
        }

        if (verbose)
            System.out.println("Point C sensor values: (" + sensorsC.getYawValue() + ", " + sensorsC.getPitchValue() + ")");

        return sensorsC;
    }

    /**
    *   Calculates the calibration data, using the 3 points set as A, B and C.
    *
    *   @return     the calculated calibration data object
    *   @exception Exception   if the underlying communications object failed to execute the method call
    *   @exception Exception  if either one of the 3 calibration points A, B or C have not yet been set
    */
    public CalibrationData calculateCalibration() throws Exception
    {
        if (verbose)
            System.out.println("Calculating calibration data...");

        if (sensorsA == null || sensorsB == null || sensorsC == null)
        {
            throw new Exception("Calibration points A, B, and C must all be set before calculating calibration parameters.");
        }

        // work out the pitch difference between points A and B
        int pitchdiff = Math.abs( sensorsB.getPitchValue() - sensorsA.getPitchValue() );
        // work out the yaw difference between points A and C
        int yawdiff = Math.abs( sensorsC.getYawValue() - sensorsA.getYawValue() );

        double pitchratio = (double)pitchdiff / Math.abs(POINTB_PITCHANGLE - POINTA_PITCHANGLE);
        double yawratio = (double)yawdiff / Math.abs(POINTC_YAWANGLE - POINTA_YAWANGLE);

        // set the calibration data
        calibdata = new CalibrationData( yawratio, pitchratio );

        if (verbose)
        {
            System.out.println("Yaw ratio: " + yawratio);
            System.out.println("Pitch ratio: " + pitchratio);
        }

        if (verbose)
            System.out.println("Moving camera to (0, 0).");

        // move camera to true 0,0
        try
        {
            comm.moveTo(0, 0);
        }
        catch (IRException e)
        {
            throw new Exception("Could not move camera back to (0,0).");
        }

        if (verbose)
        {
            System.out.println("Calibration complete.");
        }

        // return the calibration data
        return calibdata;
    }

}
