//  PCAngleLayer.java
/**
 *  Part of PC307 Group M Project: Lego RCX WebCam Mount
 *  Category: PC Control
 *
 *  This class provides a high level way of manipulating the RCX's position from the PC.
 *  It has functions that enable movement of the RCX in terms of angles.
 *  It requires calibration data, and so calibration data must be passed to the constructor. 
 *  ie. calibration should be done before creating an instance of this class.
 *
 *  @author Philip Hagen
 *
 *  Last updated: 10 October 2003
 */

public class PCAngleLayer
{

    // the layer used to communicate with the RCX
    private PCCommLayer comm;

    // the calibration data being used to convert angles to sensor values
    // (and vice versa)
    private CalibrationData calibdata;

    /**
    *   Constructor to create a PCAngleLayer, with calibration parameters as specified
    *   in the given CalibrationData argument.  The given PCCommLayer is used for
    *   communications with the PC (this may be useful as generally only one instance
    *   of the PCCommLayer can be used at a time).
    *
    *   @param      cd              the calibration data object
    *   @exception  IRException  if the underlying communications object could not be created
    */
    public PCAngleLayer(PCCommLayer pccomm, CalibrationData cd)
    {
        this.comm = pccomm;
        this.calibdata = cd;
    }

    /**
    *   Constructor to create a PCAngleLayer, with calibration parameters as specified
    *   by the given yawratio and pitchratio. The given PCCommLayer is used for
    *   communications with the PC (this may be useful as generally only one instance
    *   of the PCCommLayer can be used at a time).
    *
    *   @param      yawration       the yawratio value
    *   @param      pitchratio      the pitchratio value
    *   @exception  IRException  if the underlying communications object failed to execute the method call
    */
    public PCAngleLayer(PCCommLayer pccomm, double yawratio, double pitchratio)
    {
        this.comm = pccomm;
        this.calibdata = new CalibrationData(yawratio, pitchratio);
    }

    /**
    *   Constructor to create a PCAngleLayer, with calibration parameters as specified
    *   in the given CalibrationData argument.  A new PCCommLayer is created internally
    *   for communications with the RCX.
    *
    *   @param      cd              the calibration data object
    *   @exception  IRException  if the underlying communications object could not be created
    */
    public PCAngleLayer(CalibrationData cd) throws IRException
    {
        this.comm = new PCCommLayer();
        this.calibdata = cd;
    }

    /**
    *   Constructor to create a PCAngleLayer, with calibration parameters as specified
    *   by the given yawratio and pitchratio. A new PCCommLayer is created internally
    *   for communications with the RCX.
    *
    *   @param      yawration       the yawratio value
    *   @param      pitchratio      the pitchratio value
    *   @exception  IRException  if the underlying communications object failed to execute the method call
    */
    public PCAngleLayer(double yawratio, double pitchratio) throws IRException
    {
        this.comm = new PCCommLayer();
        this.calibdata = new CalibrationData(yawratio, pitchratio);
    }

    /**
    * Pings the RCX. Basically just sends a command to the RCX, and times how long the RCX
    * takes to reply. Used to check if the RCX is responding.
    *
    * @return      the time that it took recieve a respone from the RCX. (in milliseconds)
    * @exception   IRException     if the method fails to send a correct code,
    *                                              times out or receives an incorrect code response
    */
    public long ping() throws IRException
    {
        return comm.ping();
    }

     /**
     * Moves the camera to an absolute position using degrees.
     *
     * @param   yaw     the target yaw value in degrees
     * @param   pitch   the target pitch value in degrees
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void moveTo(int yaw, int pitch) throws IRException
    {
        comm.moveTo(
            (int)(yaw * calibdata.getYawRatio()),
            (int)(pitch * calibdata.getPitchRatio()) );
    }

    /**
     * Moves the camera to a relative position using raw sensor values.
     *
     * @param   yaw     the desired yaw value in degrees
     * @param   pitch   the desired pitch value in degrees
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void moveBy(int yaw, int pitch) throws IRException
    {
        comm.moveBy(
            (int)(yaw * calibdata.getYawRatio()),
            (int)(pitch * calibdata.getPitchRatio()) );
    }

    /**
     *  Gets the current camera position from the RCX as a camera positon object
     *  by using yaw and pitch values and ratios.
     *
     *  @return   the sensor values of the camera in the current position
     *  @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public CamPosition getPosition() throws IRException
    {
        SensorValues sv = comm.getPosition();
        return new CamPosition(
            (int)((double)sv.getYawValue() / calibdata.getYawRatio()),
            (int)((double)sv.getPitchValue() / calibdata.getPitchRatio()) );
    }

    /**
     * Sets the current camera position for the RCX using  sensor values
     * and yaw and pitch ratio values.
     *
     * @param   yaw       the yaw sensor value
     * @param   pitch     the pitch sensor value
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void setPosition(int yaw, int pitch) throws IRException
    {
        comm.setPosition( new SensorValues(
            (int)(yaw * calibdata.getYawRatio()),
            (int)(pitch * calibdata.getPitchRatio()) ) );
    }
    
    /**
     * Sets the current camera position for the RCX using a camera position data oject.
     *
     * @param   cp     the camera position object
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void setPosition(CamPosition cp) throws IRException
    {
        setPosition( cp.getYaw(), cp.getPitch() );
    }

    /**
     *  Executes the standard shut down procedure of the underlying communications layer
     *  and properly closes down all executing threads on the RCX and PC.
     *
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code respon
     */
    public void shutdown() throws IRException
    {
        comm.shutdown();
    }

    /**
     *  Returns the communications layer being used to talk to the RCX by this object.
     *  @return The instance of PCCommLayer being used.
     */
    public PCCommLayer getCommLayer()
    {
        return comm;
    }
    
    /**
     *  Returns the calibration data being used.
     */
    public CalibrationData getCalibrationData()
    {
        return calibdata;
    }

}
