// PCCommLayer.java

/**
 *  Part of PC307 Group M Project: Lego RCX WebCam Mount
 *  Category: PC
 *
 *  This class resides on the PC and has its methods called
 *  (either directly by the Matlab user or indirectly through
 *  other Java classes). Once a PCCommLayer object is
 *  created the required methods can be invoked and 
 *  communication with the RCX can be established.
 *
 *  @author Iavor Georgeff
 *  @author Philip Hagen
 *  @version %I%, %G%
 */

import java.io.*;
import josx.rcxcomm.*;

public class PCCommLayer implements PCCommInterface
{

    // ----------------------- CONSTANTS -------------------------

    // flag to turn verbose operation on or off
    public static final boolean verbose = false;

    // codes for IR communication between PC and RCX
    public static final int IRCODE_PING                     =   1;
    public static final int IRCODE_PING_RESPONSE            =   2;
    public static final int IRCODE_GETPOSITION              =   3;
    public static final int IRCODE_GETPOSITION_RESPONSE     =   4;
    public static final int IRCODE_SETPOSITION              =   5;
    public static final int IRCODE_SETPOSITION_RESPONSE     =   6;
    public static final int IRCODE_MOVETO                   =   7;
    public static final int IRCODE_MOVETO_RESPONSE          =   8;
    public static final int IRCODE_MOVEBY                   =   9;
    public static final int IRCODE_MOVEBY_RESPONSE          =   10;
    public static final int IRCODE_RCXDONE                  =   11;
    public static final int IRCODE_RCXERROR                 =   12;
    public static final int IRCODE_SHUTDOWN                 =   13;
    public static final int IRCODE_SHUTDOWN_RESPONSE        =   14;

    // RCX error codes.
    // an IRCODE_RCXERROR will be followed by one of these to indicate
    // the nature of the error
    public static final int IRERROR_INVALIDCODE                     = 1;
    public static final int IRERROR_BADPARAMCOUNT                   = 2;

    // constants for directions of movement
    public static final int DIRECTION_UP    =   1;
    public static final int DIRECTION_DOWN  =   2;
    public static final int DIRECTION_LEFT  =   3;
    public static final int DIRECTION_RIGHT =   4;

    // Some static constant definitions
    public static final int READ_TIMEOUT = 6000;

    // used when converting a double to an int for sending
    // across IR link as an int
    private static final int DOUBLE_TO_INT_RATIO = 1000;

    // -------------------- INSTANCE VARIABLES ----------------------------

    private RCXPort rcxport = null;
    private DataInputStream dis = null;
    private DataOutputStream dos = null;

    // -------------------- CLASS METHODS ---------------------------------

    /**
     * Default constructor. Attempts to initialize the input and output
     * data streams required for communication.
     *
     * @exception   IRException     if the RCXPort communication object could not be created
     */
    public PCCommLayer() throws IRException
    {
        if (verbose)
            System.out.println("Initialising PC Comm Layer...");

        try
        {
            // Create an RCXport object for communication
            // Create the required Data I/O streams
            rcxport = new RCXPort();
            dis = new DataInputStream( rcxport.getInputStream() );
            dos = new DataOutputStream( rcxport.getOutputStream() );
        }
        catch (IOException ioe)
        {
            throw new IRInitException("Could not create RCXPort object");
        }

        // This sets the timeout for reading from the data output stream
        rcxport.setTimeOut(READ_TIMEOUT);

        if(verbose)
            System.out.println("Port timeout is " + rcxport.getTimeOut() + " ms.");

    }

    /**
     * Finalize method, that is called by the garbage collector when the object is about
     * to be destroyed.  Closes the input and output streams and the rcx port.
     */
    protected void finalize()
    {
        close();
    }


    /**
     * Closes the input and output streams and the rcx port.
     */
    public void close()
    {
        try
        {
            dis.close();
            dos.close();
        }
        catch (Exception e)
        {
            System.out.println("PCCommLayer.finalize(): error closing streams.");
        }
        rcxport.close();
    }

    
    /**
     * Pings the RCX. Basically just sends a command to the RCX, and times how long the RCX
     * takes to reply. Used to check if the RCX is responding.
     *
     * @return      the time that it took receive a respone from the RCX. (in milliseconds)
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public long ping() throws IRException
    {
        if(verbose)
            System.out.println("Pinging RCX...");

        long t1, t2;
        long dt = 0;
        t1 = System.currentTimeMillis();

        try
        {
            dos.writeInt(IRCODE_PING);
        }
        catch (IOException ioe)
        {
            throw new IRSendException("Failed attempting to send: ping");
        }

        if(verbose)
            System.out.println("Awaiting response...");

        try
        {
            int a = dis.readInt();
            t2 = System.currentTimeMillis();

            if (a == IRCODE_PING_RESPONSE)
            {
                dt = t2 - t1;

                if(verbose)
                    System.out.println("Recieved response in " + dt + " ms.");
            }
            else if (a == IRCODE_RCXERROR)
            {
                handleErrorInfo();
            }
            else
            {
                throw new IRBadResponseException("RCX returned an unknown opcode: " + a);
            }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting for ping response");
        }

        return dt;
    }

    /**
     * Gets the current camera position from the RCX as a sensor values object.
     *
     *@return   the sensor values of the camera in the current position
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public SensorValues getPosition() throws IRException
    {
        if(verbose)
            System.out.println("Sending command: get camera position");

        int a,b,c;

        SensorValues sv = null;

        try
        {
            dos.writeInt(IRCODE_GETPOSITION);
        }
        catch (IOException ioe)
        {
            throw new IRSendException("Failed attempting to send: get camera position");
        }

        if(verbose)
            System.out.println("Awaiting response...");

        try
        {
            a = dis.readInt();
            if (a == IRCODE_GETPOSITION_RESPONSE)
            {
                // Should receive 2 integers as arguments
                b = dis.readInt();
                c = dis.readInt();

                if(verbose) {
                    System.out.println("Yaw sensor value is: " + b);
                    System.out.println("Pitch sensor value is: " + c);
                }

                sv = new SensorValues(b,c);
            }
            else if (a == IRCODE_RCXERROR)
            {
                handleErrorInfo();
            }
            else
            {
                throw new IRBadResponseException("RCX returned an unknown opcode: " + a);
            }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting for get position response");
        }

        return sv;
    }


    /**
     * Sets the current camera position for the RCX using raw sensor values.
     *
     * @param   yawvalue        the yaw sensor value
     * @param   pitchvalue      the pitch sensor value
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void setPosition(int yawvalue, int pitchvalue) throws IRException
    {
        setPosition( new SensorValues(yawvalue, pitchvalue) );
    }


    /**
     * Sets the current camera position for the RCX using sensor values data object.
     *
     * @param   sv     the current camera sensor values
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void setPosition(SensorValues sv) throws IRException
    {
        if(verbose)
            System.out.println("Sending command: set sensor values");

        int a;

        try
        {
            // send code
            dos.writeInt(IRCODE_SETPOSITION);
            // send yaw sensor value
            dos.writeInt(sv.getYawValue());
            // send pitch sensor value
            dos.writeInt(sv.getPitchValue());
        }
        catch (IOException ioe)
        {
            throw new IRSendException("Failed attempting to send: set sensor values");
        }

        if (verbose)
            System.out.println("Awaiting response...");

        try
        {
            a = dis.readInt();
            if (a == IRCODE_SETPOSITION_RESPONSE)
            {
                if (verbose)                    
                    System.out.println("Ok.");
            }
            else if (a == IRCODE_RCXERROR)
            {
                handleErrorInfo();
            }
            else
            {
                throw new IRBadResponseException("RCX returned an unknown opcode: " + a);
            }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting for: set sensor values response");
        }
    }


    /**
     * Moves the camera to an absolute position using raw sensor values.
     *
     * @param   yaw     the target yaw sensor value
     * @param   pitch   the target pitch sensor value
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void moveTo( int yaw, int pitch ) throws IRException
    {
        if(verbose)
        {
            System.out.println("Sending command: move to sensor values (" + yaw + ", " + pitch + ")");
        }

        try
        {
            // send code
            dos.writeInt(IRCODE_MOVETO);
            // send yaw amount to move
            dos.writeInt(yaw);
            // send pitch amount to move
            dos.writeInt(pitch);
        }
        catch (IOException ioe)
        {
            throw new IRSendException("Failed attempting to send: move to sensor values");
        }

        if(verbose)
            System.out.println("Awaiting response...");

        int a;

        try
        {
            a = dis.readInt();
            if (a == IRCODE_MOVETO_RESPONSE)
            {
                if(verbose) {
                System.out.println("Ok."); }
            }
            else if (a == IRCODE_RCXERROR)
            {
                handleErrorInfo();
            }
            else
            {
                throw new IRBadResponseException("RCX returned an unknown IR code: " + a);
            }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting for move to sensor values response");
        }

    }


    /**
     * Moves the camera to an absolute position using sensor values data object.
     *
     * @param   sv   the target camera sensor values
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void moveTo(SensorValues sv)  throws IRException
    {
        moveTo( sv.getYawValue(), sv.getPitchValue() );
    }


    /**
     * Moves the camera to a relative position using raw sensor values.
     *
     * @param   yaw     the desired yaw sensor value
     * @param   pitch   the desired pitch sensor value
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void moveBy( int yaw, int pitch ) throws IRException
    {
        if(verbose)
        {
            System.out.println("Sending command: move by sensor values (" + yaw + ", " + pitch + ")");
        }

        try
        {
            // send code
            dos.writeInt(IRCODE_MOVEBY);
            // send yaw amount to move
            dos.writeInt(yaw);
            // send pitch amount to move
            dos.writeInt(pitch);
        }
        catch (IOException ioe)
        {
            throw new IRSendException("Failed attempting to send: move by sensor values");
        }

        if(verbose)
            System.out.println("Awaiting response...");

        int a;

        try
        {
            a = dis.readInt();
            if (a == IRCODE_MOVEBY_RESPONSE)
            {
                if(verbose)
                    System.out.println("Ok.");
            }
            else if (a == IRCODE_RCXERROR)
            {
                handleErrorInfo();
            }
            else
            {
                throw new IRBadResponseException("RCX returned an unknown IR code: " + a);
            }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting for move by sensor values response");
        }

    }


    /**
     * Moves the camera to a relative position using raw sensor values.
     *
     * @param   sv      The sensor values to move by.
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void moveBy(SensorValues sv) throws IRException
    {
        moveBy( sv.getYawValue(), sv.getPitchValue() );
    }


    
    /**
     * Properly prepares the system for shut-down and stops execution of the running
     * processes on the RCX, as well as terminating all other execution threads on the PC.
     *
     * @exception   IRException     if the method fails to send a correct code,
     *                                              times out or receives an incorrect code response
     */
    public void shutdown() throws IRException
    {
        if(verbose)
            System.out.println("Sending shutdown command to RCX...");

        try
        {
            dos.writeInt( IRCODE_SHUTDOWN );
        }
        catch (IOException ioe)
        {
            throw new IRSendException("Failed attempting to send: shutdown()");
        }

        if(verbose)
            System.out.println("Awaiting response...");

        try
        {
            int a = dis.readInt();
            if (a == IRCODE_SHUTDOWN_RESPONSE)
            {
                if (verbose)
                {
                    System.out.println("RCX shutting down.");
                    System.out.println("PC software shutting down.");
                }
            }
            else if (a == IRCODE_RCXERROR)
                {
                    handleErrorInfo();
                }
            else
                {
                    throw new IRBadResponseException("RCX returned an unknown opcode: " + a);
                }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting for shutdown response");
        }

        close();
    }

    /**
    *  This method is called whenever an IRCODE_RCXERROR is received.
    *  It tries to read the next parameter which will be the type of error,
    *  and handles an appropriate message.
    *
    * @exception   IRException     if the method fails to send a correct code,
    *                                              times out or receives an incorrect code response
    */
    private void handleErrorInfo() throws IRException
    {
        int err;

        try
        {
            err = dis.readInt();

            switch(err)
            {
                case IRERROR_INVALIDCODE:
                    throw new IRRCXErrorException("RCX received invalid IR code");
                case IRERROR_BADPARAMCOUNT:
                    throw new IRRCXErrorException("RCX received incorrect number of parameters");
                default:
                    throw new IRRCXErrorException("Received an invalid RCX error parameter: " + err);
            }
        }
        catch (IOException ioe)
        {
            throw new IRReceiveException("Timed out while waiting to read RCX error code parameter");
        }
    }


}   // End of PCCommLayer class
