// WebCamCapture.java
/**
 * Part of PC307 Group M Project: Lego RCX WebCam Mount
 * Category: PC Imaging
 *
 * WebCamCapture. This class facilitates the capture of images from a webcam.
 * Methods include setup() and getCameraName(), which are integral to the setup of the camera, and
 * getBufferedImage() which returns a snapshot image.
 *
 * @author Nevin Doss 07/10/03
 * @see #WebCamCapture()
 * @see #getCameraName()
 * @see #setup(String cameraName)
 * @see #getBufferedImage()
 * @version 1.0
 */

import javax.media.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.util.Vector;
import java.util.*;
import java.util.Enumeration;
import javax.media.format.*;
import javax.media.protocol.*;
import javax.media.control.*;
import javax.media.format.YUVFormat;
import javax.media.format.VideoFormat;
import javax.media.util.BufferToImage;
import javax.imageio.*;

public class WebCamCapture extends Thread implements ControllerListener{
/**
 * The verbose flag, set to True if user wishes to view verbose messages
 */
    private static final boolean verbose = false;
/**
 * Field to set the returning of delimiters to true in StringTokenizer methods
 * @see java.util.StringTokenizer#StringTokenizer(String str, String delim, boolean returnDelims)
 */
    private final boolean RETURNDELIMS = true;
 /**
  * Field to indicate the first letter of each string
  */
    private final int FIRSTLETTER = 0;
/**
 * The Fileformat that the user wishes to use. Initially set a "jpg"
 * @see WebCamCapture#saveCurrentImage(String)
 */
    private static String FILEFORMAT = "jpg";
/**
 * This field sets the X_resolution to default 320, however this is changed by the constructor
 * @see WebCamCapture#WebCamCapture(int x_res, int y_res)
 */
    private static int DEFAULT_X_RES = 320;
    /**
    * This field set the Y_resolution to default 240, however this is changed by the constructor
    * @see WebCamCapture#WebCamCapture(int x_res, int y_res)
    */
    private static int DEFAULT_Y_RES = 240;

    private Object stateLock = new Object();
    private Object runLock = new Object();
/**
 * This field is a Global CaptureDeviceInfo object that will allow Image capture
 * @see WebCamCapture#setup(String cameraName)
 */
    private CaptureDeviceInfo cdi;
    /**
     * This field is a Global Processor object that will allow Image capture
 * @see WebCamCapture#setup(String cameraName)
     */
    private Processor deviceProc = null;
    /**
     * This field is a Global PushBufferStream object that will allow Image capture
 * @see WebCamCapture#setup(String cameraName)
     */
    private PushBufferStream camStream;
/**
 * This field is a Global PushBufferDataSource object that will facilitate Image Capture
 * @see WebCamCapture#setup(String cameraName)
 */
    private PushBufferDataSource source = null;
    private DataSource formattedSource = null;
    /**
     * This field is a Global BufferToImage object that will facilitate in image conversion
     * @see WebCamCapture#getBufferedImage()
     */
    private BufferToImage converter;
    /**
     * This is the Image object that will be written to and read from to generate the host of images that will constitute the image capture
     * @see WebCamCapture#getImage()
     */
    private Image currentImage;
    private boolean threadRunning = false;
    private boolean quitrequest = false;

    /**
     * The Basic WebCamCapture constructor, creates a WebCamCapture object and
     * sets all variables to default
     * @see #WebCamCapture(int, int)
     * @see #WebCamCapture(String)
     * @see #WebCamCapture(int, int, String)
     **/
    public WebCamCapture(){
    }

    /**
     * This constructor sets the Default X resolution and Default Y resolution to user defined inputs
     * This can be found by running the printYUVDevices() method
     *  @param x_res    The user defined X-Resolution
     *  @param y_res    The user defined Y-Resolution
     *  @see #printYUVDevices()
     *  @see WebCamCapture#DEFAULT_X_RES
     *  @see WebCamCapture#DEFAULT_Y_RES
     *
     **/
    public WebCamCapture(int x_res, int y_res){
        this.DEFAULT_X_RES = x_res;
        this.DEFAULT_Y_RES = y_res;
    }

    /**
     * This constructor sets the Default file saving format to a user defined value, which can be selected from a list
     * given in ImageIO api
     * @param fformat   The user defined File Format output
     *
     **/
    public WebCamCapture(String fformat){
        this.FILEFORMAT = fformat;
    }

    /**
     * This constructor sets the Default X resolution and Default Y resolution to user defined inputs
     * aswell as setting the default file saving format.
     * @param x_res     The user defined X-Resolution
     * @param y_res     The user defined Y-Resolution
     * @param fformat   The user defined File Format output
     **/
    public WebCamCapture(int x_res, int y_res, String fformat){
        this.DEFAULT_X_RES = x_res;
        this.DEFAULT_Y_RES = y_res;
        this.FILEFORMAT = fformat;
    }

    /**
     * A method designed to give the user an idea what types of formats the current
     * Video Capture device is capable of handling
     * prints this out on the default System.out stream.
     **/
    public void printYUVDevices(){
        YUVFormat yuvf = new YUVFormat();
        Vector videoDevs = CaptureDeviceManager.getDeviceList(yuvf);

        for(int i = 0; i < videoDevs.size(); i++){
            System.out.println("\n" + videoDevs.elementAt(i).toString());
        }
    }

    /**
     * A method designed to return the Camera Name of the capture device.
     * By assuming that all camera capture devices have the first letter
     * of their string as a v.
     * @return The camera name of the device currently connected to the computer
     * @throws WebCamCaptureException Thrown when could not find video device
     **/
    public String getCameraName() throws WebCamCaptureException{
        YUVFormat yuvf = new YUVFormat();
        Vector videoDevs = CaptureDeviceManager.getDeviceList(yuvf);
        String cameraName = null;
        StringTokenizer myToken = null;
        for(int i = 0; i < videoDevs.size(); i++){

            myToken = new StringTokenizer(videoDevs.elementAt(i).toString(), ":", RETURNDELIMS);
            cameraName = myToken.nextToken();

            if(cameraName.charAt(FIRSTLETTER) == 'v')
                break;
            else if(i == (videoDevs.size() - 1)){
	    	if(verbose)
                	System.out.println("Could not find video device, please reconnect and restart");
                throw new WebCamCaptureException("Could not find video device, please reconnect and restart");
            }
        }

        for(int i = 0; i < 4; i++){
            cameraName = cameraName + myToken.nextToken();
            //System.out.println(i+" " + cameraName);
        }

    /* By knowing the the length of string "v41:Logitech QuickCam Pro 3000:0" is 32
     * We know the the tokenizer has added the extra " " after the last "0"
     * hence we have to chop off the last " ", if not CaptureDeviceManager will not
     * accept the string as a valid camera Name
     */
        cameraName = cameraName.substring(FIRSTLETTER, (cameraName.length() - 1));
        //System.out.println("|"+cameraName+"|");
        return cameraName;

    }

    /**
     * Method that will setup the CaptureDevice by going through appropriate steps
     * and by creating required objects. This method must be run before any methods
     * that require a working CaptureDeviceInfor object and a MediaLocator.
     * Steps that this method will go through are:
     * 1. Create a new CaptureDeviceInfo.
     * 2. Sets the format that is by defaul or by user definition.
     * 3. Sets the MediaLocator for the current device.
     * 4. Assigns a DataSource for the MediaLocator.
     * 5. Connects to the DataSource.
     * 6. Sets up a Processor and attempts to realize the Processor.
     * 7. Starts the processor - Pushing the datastream onto the buffer.
     * @param cameraName
     * @throws WebCamCaptureException
     **/
    public void setup(String cameraName) throws WebCamCaptureException{
        CaptureDeviceInfo device = CaptureDeviceManager.getDevice(this.getCameraName());

        if(device == null)
            throw new WebCamCaptureException("No device Found [" + this.getCameraName() + "]");

        YUVFormat userFormat = null;
        Format[] cfmt = device.getFormats();
        Dimension viewSize = new Dimension(DEFAULT_X_RES, DEFAULT_Y_RES);

        /* Find the format that the user has requested (if available) */

        for(int i = 0; i < cfmt.length ; i++){
            if(cfmt[i] instanceof YUVFormat){
                userFormat = (YUVFormat)cfmt[i];
                Dimension d = userFormat.getSize();

                if(viewSize.equals(d))
                    break;

                userFormat = null;
            }
        }

        if(userFormat == null){
            System.out.println("Cannot find Format, please enter new formats and recompile");
            throw new WebCamCaptureException("Cannot find Format, please enter new formats and recompile");
        }

        MediaLocator loc = device.getLocator();

        if(loc == null){
            throw new WebCamCaptureException("Unable to get MediaLocator for device");
        }


        /* New create a dataSource for this device and set the format to the one chosen by the user */

        try{
            formattedSource = Manager.createDataSource(loc);
        } catch (IOException ioe){
            throw new WebCamCaptureException("IO Error creating dataSource");
        } catch (NoDataSourceException ndse){
            throw new WebCamCaptureException("Unable to create DataSource");
        }

        /* This is the method to set the format */

        if(!(formattedSource instanceof CaptureDevice))
            throw new WebCamCaptureException("DatSource not a CaptureDevice");

        FormatControl[] fmtControls = ((CaptureDevice)formattedSource).getFormatControls();

        if(fmtControls == null || fmtControls.length == 0)
            throw new WebCamCaptureException("No FormatControl available");

        Format setFormat = null;


        /* We need to loop throught the available FormatControls and try to set the format to the one we want */

        for(int i = 0; i < fmtControls.length; i++){
            if(fmtControls[i] == null)
                continue;

            if((setFormat = fmtControls[i].setFormat(userFormat)) != null)
                break;
        }

        /* Throw an exception if we couldn't set the format */

        if(setFormat == null){
            throw new WebCamCaptureException("Failed to set camera Format!");
        }

        /* Connect to the Datasource */

        try{
            formattedSource.connect();
        } catch (IOException ioe){
            throw new WebCamCaptureException("Unable to connect to DataSource");
        }

      /* We don't want to display the output to the user at this stage, therefore using a
       * processor rather than a player */

        try{
            deviceProc = Manager.createProcessor(formattedSource);
        } catch (IOException ioe){
            ioe.printStackTrace();
        } catch (NoProcessorException npe){
            npe.printStackTrace();
        }

        deviceProc.addControllerListener(this);
        deviceProc.realize();

        while(deviceProc.getState() != Controller.Realized){
            synchronized (stateLock) {
                try {
                    stateLock.wait();
                } catch (InterruptedException ie){
                    throw new WebCamCaptureException("Failed to get into the realized state");
                }
            }
        }

        deviceProc.start();

        try {
            source = (PushBufferDataSource)deviceProc.getDataOutput();
        } catch (NotRealizedError nre) {
            /* Should never happen */
            throw new WebCamCaptureException("Processor not realized");
        }

        PushBufferStream[] streams = source.getStreams();
        camStream = null;

        for(int i = 0; i <streams.length ; i++){
            if(streams[i].getFormat() instanceof YUVFormat){
                camStream = streams[i];
                YUVFormat yuvf = (YUVFormat)streams[i].getFormat();
                converter = new BufferToImage(yuvf);
                break;
            }
        }

        if (verbose)
            System.out.println("Capture Device ready!");
    }

    /**
     * Get an Image from the camera as an AWT Image Object
     * @return Image
     **/
    public Image getImage(){
        //  System.out.println("Calling getImage()");
        while(threadRunning == false) {
            if(verbose)
                    System.out.println("threadRunning == False");
                synchronized (runLock){
                if(verbose)
                    System.out.println("synchronising runLock");
                    try{
                        runLock.wait();
                if(verbose)
                        System.out.println("RunLock has waited");
                    } catch (InterruptedException ie) {
                        //Ignore
                    }
                }
            }
        //System.out.println("I'm returning");
        return accessInternalImage(null);
    }

    /**
     * Get an image from the camera as a BufferedImage
     * @return BufferedImage
     * @see #getImage()
     **/
    public BufferedImage getBufferedImage(){
        return (BufferedImage)getImage();
    }


    /**
     * Returns an array representing the snapshot image in in array format.
     * This is designed for use with Matlab, which uses an array
     * representation for images.
     *
     * @return  A (width x height x no of colors) array representing the image.
     */
    public int[][][] getRGBArray()
    {
        // image to store snapshot
        BufferedImage bi = getBufferedImage();
        // get dimensions of image
        int width = bi.getWidth();
        int height = bi.getHeight();
        // create array to return
        int[][][] rgb = new int[height][width][3];
        // loop through each pixel (x,y) in the image
        for (int y=0; y<height; y++)
        {
            for (int x=0; x<width; x++)
            {
                // get the RGBA value at current (x,y)
                int rgbval = bi.getRGB(x,y);
                // mask out appropriate bits to obtain red,green,blue components,
                // and shift right if neccessary.
                int red = (rgbval & 0x00ff0000) >> 16;
                int green = (rgbval & 0x0000ff00) >> 8;
                int blue = rgbval & 0x000000ff;
                // set output array
                rgb[y][x][0] = red;
                rgb[y][x][1] = green;
                rgb[y][x][2] = blue;
            }
        }
        // return the array
        return rgb;
    }


    /**
     * Run method for Thread Class.
     */
    public void run(){
        if (verbose)
            System.out.println("WebCamCapture: Capture Thread starting...");
        
        Buffer b = new Buffer();

        while(!quitrequest){
            try{
                camStream.read(b);
            } catch (Exception e) {
                // Ignore nothing we can do about it
            }
            //this.yield();
            Image i = converter.createImage(b);
            accessInternalImage(i);

            if(!threadRunning){
                threadRunning = true;

                synchronized (runLock) {
                    runLock.notifyAll();
                }
            }
        }
    }

    /*
     * Method that controls access to the global image variable
     * @param image The image object that is passed to it
     * @return Image The resulting image object, when not being written too, so that the image consumer can read it.
     *
     **/

    private synchronized Image accessInternalImage(Image image){
        if(image == null){
            return currentImage;
        }
        currentImage = image;
        return null;
    }

    /* No need to include the actual format, interface the .jpg or .gif or .png
     * however you must include it into the declaration on the
     * top */

    /**
     * Method that saves a BufferedImage returned by getBufferedImage to a file, format is specified by the static field FILEFORMAT
     * @param fileName File name in the current directory
     * @see #FILEFORMAT
     **/
    public void saveCurrentImage(String fileName){
        BufferedImage bi = this.getBufferedImage();
        try {
            File file = new File(fileName + "." + this.FILEFORMAT);
            ImageIO.write(bi, FILEFORMAT, file);
            if (verbose)
                System.out.println("WebCamCapture: wrote to " + FILEFORMAT + " format");
        } catch (IOException ioe){
            ioe.printStackTrace();
        }
    }

    public String getFileFormat()
    {
        return FILEFORMAT;
    }

    /**
     * Method that controls the state of the Processor
     * @param javax.media.ControllerEvent
     **/
    public void controllerUpdate(javax.media.ControllerEvent controllerEvent) {
        if(controllerEvent instanceof RealizeCompleteEvent){
            synchronized (stateLock){
                stateLock.notifyAll();
            }
        }
    }

    public void quit()
    {
        if (verbose)
            System.out.println("WebCamCapture: quitting");
        quitrequest = true;;
        this.source.disconnect();
        this.formattedSource.disconnect();
        this.deviceProc.close();
        this.deviceProc.deallocate();
    }

    protected void finalize()
    {
        this.quit();
    }

}



