// RCXCommunications.java
/**
 *  Part of PC307 Group M Project: Lego RCX WebCam Mount
 *  Category: RCX
 *
 *  This class handles communications between the PC and the RCX.
 *  Implements the IRCommunicationsInterface interface.
 *  Uses an RCXPort object and streams to send/recieve data.
 *
 *  @author Philip Hagen    hagenp01@tartarus.uwa.edu.au
 *
 *  Last updated:   10 October 2003
 */

import java.io.*;
import josx.rcxcomm.*;
import josx.platform.rcx.*;

public class RCXCommunicationsLayer implements RCXCommunicationsInterface
{

    // flag to turn verbose operation (ie. printing to LCD) on or off
    public static final boolean verbose = true;

    // codes for IR communication between PC and RCX
    public static final int IRCODE_PING                     =   1;
    public static final int IRCODE_PING_RESPONSE            =   2;
    public static final int IRCODE_GETPOSITION              =   3;
    public static final int IRCODE_GETPOSITION_RESPONSE     =   4;
    public static final int IRCODE_SETPOSITION              =   5;
    public static final int IRCODE_SETPOSITION_RESPONSE     =   6;
    public static final int IRCODE_MOVETO                   =   7;
    public static final int IRCODE_MOVETO_RESPONSE          =   8;
    public static final int IRCODE_MOVEBY                   =   9;
    public static final int IRCODE_MOVEBY_RESPONSE          =   10;
    public static final int IRCODE_RCXDONE                  =   11;
    public static final int IRCODE_RCXERROR                 =   12;
    public static final int IRCODE_SHUTDOWN                 =   13;
    public static final int IRCODE_SHUTDOWN_RESPONSE        =   14;

    // RCX error codes.
    // an IRCODE_RCXERROR will be followed by one of these to indicate
    // the nature of the error
    public static final int IRERROR_INVALIDCODE                     = 1;
    public static final int IRERROR_BADPARAMCOUNT                   = 2;

    // other constant definitions
    public static final int READ_TIMEOUT = 500;

    // member variables
    private RCXPort rcxport = null;
    private DataInputStream dis = null;
    private DataOutputStream dos = null;

    // the message to be sent (back) to the PC
    private IRMessage outmsg;

    /**
     * Default constructor. Attempts to initialize the input and output
     * data streams required for communication.
     */
    public RCXCommunicationsLayer()
    {
        // try to create an RCXPort object, then create a
        // DataInputStream and DataOutputStream from it
        try
        {
            rcxport = new RCXPort();
            dis = new DataInputStream( rcxport.getInputStream() );
            dos = new DataOutputStream( rcxport.getOutputStream() );
            rcxport.setTimeOut( READ_TIMEOUT );
        }
        catch (IOException e)
        {
            // handle error
            debugPrint("ERR1");
            System.exit(1);
        }

        // initialise the IR message
        outmsg = new IRMessage();
    }


    /**
    *   Sends the given message across the IR link to the PC.
     *  Note that the underlying RCXPort uses a LLCReliableHandler to send
     *  the data, and this keeps trying to send until it succeeds, which means
     *  that this method can block for an indefinite amount of time.
     *
     *  @param      msg     the IRMessage object to send to the PC.
     *  @return    true     if the send was successful, or false if it failed.
     */
    public boolean sendMessage(IRMessage msg)
    {
        if (msg == null)
        {
            // quit with error
            debugPrint("E C1");
            System.exit(1);
        }

        try
        {
            dos.writeInt( msg.getCode() );
            for (int i=0; i<msg.getParamCount(); i++)
                dos.writeInt( msg.getParam(i) );
        }
        catch (IOException e)
        {
            debugPrint("E SND");
            System.exit(1);
            return false;
        }
        return true;
    }


    /**
     *  Checks for any incoming messages (sent from the PC).
     *  If there is a message available, then the message parameter
     *  is set appropriately.
     *  If no message was available, then the message parameter has
     *  its code set to -1.
     *  <p>
     *  This method attempts to read a code from the DataInputStream.
     *  Then from the code it knows how many parameters to expect for
     *  the message, and tries to read this many from the stream.
     *  Once all parameters have been recieved, it creates an IRMessage
     *  and stores it in the class variable msg.
     *
     *  @param msg  the message object in which to store the incoming
     *              message (if there was one).  Must be an initialised
     *              IRMessage object (ie. not null).  Will have its code
     *              set to -1 if no message was available.
     */
    public void getMessage(IRMessage msg)
    {
        // try to read from the IR port to see if there is any data there...
        int code;
        try
        {
            code = dis.readInt();
        }
        catch (IOException e)
        {
            // this means that the read timed out.
            // in other words, there was nothing there to read, so just return null.
            msg.set(-1);
            return;
        }

        // for debugging:  beep and print code of recieved message on LCD
        if (verbose)
        {
            TextLCD.print("MSG");
        }

        // see if the integer recieved was a valid IR code, and if so
        // create an IRMessage object representing the message and its
        // parameters.

        // if the message had no parameters just create a message with just the code
        // and no parameters
        if ( (code == IRCODE_PING)  || (code == IRCODE_GETPOSITION) ||
            (code == IRCODE_GETPOSITION) || (code == IRCODE_SHUTDOWN) )
        {
            msg.set(code);
            return;
        }
        // if the message had 2 parameters, read the next 2 parameters from the buffer,
        // create a message, and return it
        else if ( (code == IRCODE_SETPOSITION) ||
            (code == IRCODE_MOVETO) || (code == IRCODE_MOVEBY) )
        {
            // try to read the 2 parameters
            int p1, p2;
            try
            {
                p1 = dis.readInt();
                p2 = dis.readInt();
                msg.set(code, p1, p2);
                return;
            }
            catch (IOException e)
            {
                // handle exception
                // what do we do... return null?  send back an error code?
                // thrown an exception that can be handled by RCXManager?

                outmsg.set(IRCODE_RCXERROR, IRERROR_BADPARAMCOUNT);
                sendMessage( outmsg );
                debugPrint("E PRM");

                System.exit(1);
            }

        }
        // the code was invalid
        else
        {
            // do we handle this here, or leave it up to RCXManager???
            outmsg.set(IRCODE_RCXERROR, IRERROR_INVALIDCODE);
            sendMessage(outmsg);
            debugPrint("E COD");
            System.exit(1);
        }

    }


    /** Prints the given string to the RCX's LCD screen, and pauses
     *  for a short time so that the user can read it.
     *  Useful for debugging.
     */
    private static void debugPrint(String msg)
    {
        if (verbose)
        {
            TextLCD.print(msg);
            for (int i=0; i<10000; i++);
        }
    }

}
