// RCXHardwareLayer.java
/**
 *  Part of PC307 Group M Project: Lego RCX WebCam Mount
 *  Category: RCX
 *
 *  This class is used for any interactions with the hardware
 *  such as moving the camera mount, getting its current position,
 *  etc.  Its methods are called from RCXManager.
 *
 *  @author Tim Malloch
 *  @author Philip Hagen
 *
 *  Last updated: 10 October 2003
 */


import josx.platform.rcx.*;


public class RCXHardwareLayer implements RCXHardwareInterface
{

    // flag to turn verbose operation (ie. printing to LCD) on or off
    public static final boolean verbose = true;

    // constants for directions of movement
    public static final int DIRECTION_UP    =   1;
    public static final int DIRECTION_DOWN  =   2;
    public static final int DIRECTION_LEFT  =   3;
    public static final int DIRECTION_RIGHT =   4;

    // set up which motor controls yaw and which controls pitch
    private static Motor YawMotor       =   Motor.A;
    private static Motor PitchMotor     =   Motor.B;

    // set up which sensor measures yaw and which measures pitch
    private static Sensor YawSensor     =   Sensor.S1;
    private static Sensor PitchSensor   =   Sensor.S2;

    // the threshold for movement
    private static final int THRESHOLD = 2;
    private static final int POWER_THRESHOLD = 10;

    // power levels for motors
    private static final int POWER_HIGH = 7;
    private static final int POWER_LOW = 2;

    // constants used as return values
    private static final int SUCCESS = 1;
    private static final int FAILURE = 0;

    // max and min extent of motion (in sensor vals)
    // these will have to be set using a function, once calibration is done
    // yaw: -180 to 180 degrees
    // pitch: -90 to 80 degrees
    // yawratio/pitchratio should be (2.166, 2.166)
    private int YAW_MIN = -390;
    private int YAW_MAX = 390;
    private int PITCH_MIN = -195;
    private int PITCH_MAX = 195;


    /** Constructor.
     */
    public RCXHardwareLayer()
    {
        // register the sensors
        YawSensor.setTypeAndMode(
            SensorConstants.SENSOR_TYPE_ROT,
            SensorConstants.SENSOR_MODE_ANGLE );
        PitchSensor.setTypeAndMode(
            SensorConstants.SENSOR_TYPE_ROT,
            SensorConstants.SENSOR_MODE_ANGLE );

        // activate the sensors
        YawSensor.activate();
        PitchSensor.activate();

    }

    /** Returns the current pitch sensor value.
     */
    public int getPitch()
    {
        return PitchSensor.readValue();
    }

    /** Returns the current yaw sensor value.
     */
    public int getYaw()
    {
        return YawSensor.readValue();
    }

    /** Resets the yaw and pitch sensors to have the given values.
     */
    public void setPosition(int yaw, int pitch)
    {
        YawSensor.setPreviousValue(yaw);
        PitchSensor.setPreviousValue(pitch);
    }

    /** Sets the bounds of the range of movement of the camera.
        Not used at present.
     */
    public void setBounds(int ymin, int ymax, int pmin, int pmax)
    {
        YAW_MIN = ymin;
        YAW_MAX = ymax;
        PITCH_MIN = pmin;
        PITCH_MAX = pmax;
    }

    /** Make the camera move to the given yaw and pitch sensor values.
     *
     *  @param  newyaw      The destination yaw value to move to.
     *  @param  newpitch    The destination pitch value to move to.
     */
    public int moveTo(int newyaw, int newpitch)
    {
        // get the current yaw and pitch, before starting to move
        int oldyaw = YawSensor.readValue();
        int oldpitch = PitchSensor.readValue();

        // check if the yaw value is out of bounds, and if so cap it
        if (newyaw > YAW_MAX)
            newyaw = YAW_MAX;
        else if (newyaw < YAW_MIN)
            newyaw = YAW_MIN;

        // check if the pitch value is out of bounds, and if so cap it
        if (newpitch > PITCH_MAX)
            newpitch = PITCH_MAX;
        else if (newpitch < PITCH_MIN)
            newpitch = PITCH_MIN;

        // set motor powers to high speed
        YawMotor.setPower(POWER_HIGH);
        PitchMotor.setPower(POWER_HIGH);
        
        // move yaw motor forwards or backwards
        if (newyaw - oldyaw > THRESHOLD)
        {
            YawMotor.forward();
        }
        else if (newyaw - oldyaw < -THRESHOLD)
        {
            YawMotor.backward();
        }

        // move pitch motor forwards or backwards
        if (oldpitch - newpitch > THRESHOLD)
        {
            PitchMotor.backward();
        }
        else if (oldpitch - newpitch < -THRESHOLD)
        {
            PitchMotor.forward();
        }

        // while either the yaw or pitch motor is on, keep looping...
        while( YawMotor.isMoving() || PitchMotor.isMoving() )
        {
            // if the difference between the current and target yaw is less than the threshold,
            // turn off the yaw motor.
            if (YawMotor.isMoving())
            {
                if (abs(YawSensor.readValue() - newyaw) < THRESHOLD)
                    YawMotor.stop();
                else if ( abs(YawSensor.readValue() - newyaw) <  POWER_THRESHOLD)
                    YawMotor.setPower(POWER_LOW);
            }
            // if the difference between the current and target pitch is less than the threshold,
            // turn off the pitch motor.
            if (PitchMotor.isMoving())
            {
                if (abs(PitchSensor.readValue() - newpitch) < THRESHOLD)
                    PitchMotor.stop();
                else if ( abs(PitchSensor.readValue() - newpitch) < POWER_THRESHOLD )
                    PitchMotor.setPower(POWER_LOW);
            }
        }

        return SUCCESS;
    }


    /** Move the camera mount slightly.  This method results in a call to moveTo().
     *
     *  @param  direction   The direction in which to move the camera, which should be one of
     *                      DIRECTION_UP, DIRECTION_DOWN, DIRECTION_LEFT, DIRECTION_RIGHT.
     *  @param  dist        The distance (in sensor values) to move the camera.
     */
    public int moveSlightly(int direction, int dist)
    {
        // get initial sensor values
        int yaw1 = YawSensor.readValue();
        int pitch1 = PitchSensor.readValue();

        if (direction == DIRECTION_UP)
        {
            TextLCD.print("UP");
            moveTo( yaw1, pitch1 + dist );
        }
        else if (direction == DIRECTION_DOWN)
        {
            TextLCD.print("DOWN");
            moveTo( yaw1, pitch1 - dist );
        }
        else if (direction == DIRECTION_LEFT)
        {
            TextLCD.print("LEFT");
            moveTo( yaw1 - dist, pitch1 );
        }
        else if (direction == DIRECTION_RIGHT)
        {
            TextLCD.print("RIGHT");
            moveTo( yaw1 + dist, pitch1 );
        }

        return SUCCESS;
    }


    /* Absolute value function. */
    private static int abs(int x)
    {
        if (x >= 0)
            return x;
        return -x;
    }


    /** Prints the given string to the RCX's LCD screen, and pauses
     *  for a short time so that the user can read it.
     *  Useful for debugging.
     */
    private static void debugPrint(String msg)
    {
        if (verbose)
        {
            TextLCD.print(msg);
            for (int i=0; i<10000; i++);
        }
    }

}
