// RCXManager.java
/**
 *  Part of PC307 Group M Project: Lego RCX WebCam Mount
 *  Category: RCX
 *
 *  This is the main top-level class that sits on the RCX and
 *  manages everything.  It uses an instance of RCXCommunicationsLayer
 *  for PC communications, and an instance of RCXHardwareLayer for
 *  interacting with the hardware.
 *
 *  @author Philip Hagen    hagenp01@tartarus.uwa.edu.au
 *
 *  Last updated:   10 October 2003
 */


import josx.platform.rcx.TextLCD;
import josx.platform.rcx.Sound;

public class RCXManager
{

    // flag to turn verbose operation (ie. printing to LCD) on or off
    public static final boolean verbose = true;

    // used when converting a double to an int for sending
    // across IR link as an int
    private static final int DOUBLE_TO_INT_RATIO = 1000;

    // default string that is displayed when the RCX is ready
    private static final String READYSTRING = "READY";

    // layer to handle communications to PC via IR port
    private RCXCommunicationsLayer comm;

    // layer to talk to hardware (motors & sensors)
    private RCXHardwareInterface hardware;

    // flag that indicates whether the program should shut down
    private boolean shutdownrequest;

    // message used to send
    private IRMessage inmsg;
    private IRMessage outmsg;


    /** 
    *   Default constructor for the RCXManager.
    *   Attempts to initialize all communication streams.
    */
    RCXManager()
    {
        TextLCD.print("HELLO");

        // create communications layer and hardware layer
        comm = new RCXCommunicationsLayer();

        hardware = new RCXHardwareLayer();
        hardware.setPosition(0, 0);
        
        shutdownrequest = false;

        inmsg = new IRMessage();
        outmsg = new IRMessage();

        // the main loop
        do
        {
            // get message from comm layer
            comm.getMessage(inmsg);

            // check if there actually was a message
            if (inmsg.getCode() != -1)
            {
                Sound.beep();
                // process the message
                processMessage();
            }

        } while (!shutdownrequest);

        debugPrint("BYE");

        // end of program
    }


    /** 
    *   Processes the given message, sends the appropriate response messsage
    *   back, and takes the appropriate actions.
    */
    private void processMessage()
    {
        // see what the message is
        // big if..else to handle the message
        if (inmsg.getCode() == comm.IRCODE_PING)
        {
            TextLCD.print("PING");
            // message is a ping so just send a confirmation
            outmsg.set( comm.IRCODE_PING_RESPONSE );
            comm.sendMessage( outmsg );
            TextLCD.print(READYSTRING);
        }
        else if (inmsg.getCode() == comm.IRCODE_MOVETO)
        {
            // message is to move to the given sensor values
            TextLCD.print("MOVTO");
            // send response
            outmsg.set(comm.IRCODE_MOVETO_RESPONSE);
            comm.sendMessage(outmsg);
            // call moveTo()
            hardware.moveTo(inmsg.getParam(0), inmsg.getParam(1));
            TextLCD.print(READYSTRING);
        }
        else if (inmsg.getCode() == comm.IRCODE_MOVEBY)
        {
            // message is to move relatively by the given yaw and pitch
            TextLCD.print("MOVBY");
            // send response
            outmsg.set(comm.IRCODE_MOVEBY_RESPONSE);
            comm.sendMessage(outmsg);
            // call moveTo()
            hardware.moveTo(
                hardware.getYaw() + inmsg.getParam(0),
                hardware.getPitch() + inmsg.getParam(1) );
            TextLCD.print(READYSTRING);
        }
        else if (inmsg.getCode() == comm.IRCODE_GETPOSITION)
        {
            // message is to send current position
            TextLCD.print("GTPOS");
            // create message with yaw and pitch sensor values
            outmsg.set(
                comm.IRCODE_GETPOSITION_RESPONSE,
                hardware.getYaw(),
                hardware.getPitch() );
            // send the message
            comm.sendMessage(outmsg);
            TextLCD.print(READYSTRING);
        }
        else if (inmsg.getCode() == comm.IRCODE_SETPOSITION)
        {
            TextLCD.print("STPOS");
            // create response message
            outmsg.set( comm.IRCODE_SETPOSITION_RESPONSE );
            // send message
            comm.sendMessage(outmsg);
            // set hardware sensor values
            hardware.setPosition( inmsg.getParam(0), inmsg.getParam(1) );
            TextLCD.print(READYSTRING);
        }
        else if (inmsg.getCode() == comm.IRCODE_SHUTDOWN)
        {
            // message was to shut down the RCX
            TextLCD.print("BYE");
            // send confirmation message
            outmsg.set(comm.IRCODE_SHUTDOWN_RESPONSE);
            comm.sendMessage(outmsg);
            // set shutdownrequest to true which will cause the main loop to terminate
            shutdownrequest = true;
        }
        else
        {
            // unrecognised opcode
            TextLCD.print("BADCD");
            outmsg.set(comm.IRCODE_RCXERROR, comm.IRERROR_INVALIDCODE);
            comm.sendMessage(outmsg);
        }
    }


    /** Prints the given string to the RCX's LCD screen, and pauses
     *  for a short time so that the user can read it.
     *  Useful for debugging.
     */
    private static void debugPrint(String msg)
    {
        if (verbose)
        {
            TextLCD.print(msg);
            for (int i=0; i<10000; i++);
        }
    }

    
    /** 
    *   This is the main method and
     *  the entry point for the program on the RCX.
     */
    public static void main(String[] args)
    {
        RCXManager rcxman = new RCXManager();
    }

}
